# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

from go_vendor_tools.cli import utils


def test_color_default() -> None:
    assert utils.color_default() is None
    assert utils.color_default({"NO_COLOR": "", "FORCE_COLOR": ""}) is None
    assert utils.color_default({"NO_COLOR": "1"}) is False
    assert utils.color_default({"FORCE_COLOR": "1"}) is True



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Shared CLI utilities
"""

from __future__ import annotations

import os
import sys
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import Any, cast

try:
    import tomlkit
except ImportError:
    HAS_TOMLKIT = False
else:
    HAS_TOMLKIT = True

from go_vendor_tools.config.utils import get_envvar_boolean
from go_vendor_tools.exceptions import MissingDependencyError, VendorToolsError


def need_tomlkit(action="this action"):
    if not HAS_TOMLKIT:
        message = f"tomlkit is required for {action}. Please install it!"
        raise MissingDependencyError(message)


def tomlkit_dump(obj: Any, path: Path) -> None:
    need_tomlkit()
    with path.open("w") as fp:
        tomlkit.dump(obj, fp)


def load_tomlkit_if_exists(path: Path | None) -> tomlkit.TOMLDocument:
    if path and path.is_file():
        with path.open("r", encoding="utf-8") as fp:
            loaded = tomlkit.load(fp)
    else:
        loaded = tomlkit.document()
    return loaded


def color_default(environ: dict[str, str] | None = None) -> bool | None:
    environ = cast(dict[str, str], os.environ if environ is None else environ)
    if environ.get("FORCE_COLOR"):
        return True
    if environ.get("NO_COLOR"):
        return False
    return None


@contextmanager
def catch_vendor_tools_error() -> Iterator[None]:
    try:
        yield
    except VendorToolsError as exc:
        if get_envvar_boolean("_GVT_DEBUG", False):
            raise
        else:
            sys.exit(f"{type(exc).__name__}: {exc}")



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

import re
import sys
from io import StringIO
from pathlib import Path
from shutil import copy2
from textwrap import dedent

import pytest
from pytest_mock import MockerFixture

from go_vendor_tools.cli import go_vendor_license, utils
from go_vendor_tools.config.base import BaseConfig
from go_vendor_tools.exceptions import MissingDependencyError
from go_vendor_tools.license_detection.base import (
    LicenseData,
    LicenseDetectorNotAvailableError,
)

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


def test_need_tomlkit(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(utils, "HAS_TOMLKIT", False)
    with pytest.raises(
        MissingDependencyError,
        match=re.escape("tomlkit is required for this action. Please install it!"),
    ):
        go_vendor_license.need_tomlkit()


def test_choose_license_detector_error_1(
    monkeypatch: pytest.MonkeyPatch, config1: BaseConfig
) -> None:
    monkeypatch.setattr(
        "go_vendor_tools.license_detection.scancode.HAS_SCANCODE", False
    )
    with pytest.raises(
        SystemExit,
        match="Failed to get detector 'scancode':"
        " The scancode-toolkit library must be installed!",
    ):
        go_vendor_license.choose_license_detector(
            "scancode", config1["licensing"], None
        )


def test_choose_license_detector_error_2(
    mocker: MockerFixture, capsys: pytest.CaptureFixture, config1: BaseConfig
) -> None:
    return_value: tuple[dict, dict] = (
        {},
        {
            "abcd": LicenseDetectorNotAvailableError("acbd is missing!?!?"),
            "123": LicenseDetectorNotAvailableError("123 is missing."),
        },
    )
    gd_mock = mocker.patch(
        "go_vendor_tools.cli.go_vendor_license.get_detectors",
        return_value=return_value,
    )
    with pytest.raises(SystemExit, match="1"):
        go_vendor_license.choose_license_detector(None, config1["licensing"], None)
    out, err = capsys.readouterr()
    assert err == "Failed to load license detectors:\n"
    expected = """\
    ! abcd: acbd is missing!?!?
    ! 123: 123 is missing.
    """
    assert dedent(expected) == out
    gd_mock.assert_called_once()


def test_red() -> None:
    with StringIO() as stream:
        go_vendor_license.red("This is an error", file=stream)
        value = stream.getvalue()
    assert value == "This is an error\n"
    with StringIO() as stream:
        stream.isatty = lambda: True  # type: ignore
        go_vendor_license.red("This is an error", file=stream)
        value = stream.getvalue()
    assert value == "\033[31mThis is an error\033[0m\n"


def test_print_licenses_all(capsys: pytest.CaptureFixture) -> None:
    directory = Path("/does-not-exist")
    license_data = LicenseData(
        directory=directory,
        license_map={
            Path("LICENSE.md"): "MIT",
            Path("LICENSE.unknown"): "Unknown",
            Path("vendor/xyz/COPYING"): "GPL-3.0-only",
        },
        undetected_licenses=frozenset(
            {
                Path("LICENSE.undetected"),
                Path("vendor/123/COPYING.123"),
            }
        ),
        unmatched_manual_licenses=(
            Path("LICENSE-Custom"),
            Path("vendor/custom/LICENSE"),
        ),
        extra_license_files=(),
        detector_name="",
    )
    go_vendor_license.print_licenses(
        results=license_data,
        unlicensed_mods=[
            Path("LICENSE.unmatched"),
            Path("vendor/123/456/LICENSE.unmatched1"),
        ],
        mode="all",
        show_undetected=True,
        show_unlicensed=True,
        directory=directory,
    )
    out, err = capsys.readouterr()
    assert not err
    expected = """\
    LICENSE.md: MIT
    LICENSE.unknown: Unknown
    vendor/xyz/COPYING: GPL-3.0-only

    The following license files were found but the correct license identifier couldn't be determined:
    - LICENSE.undetected
    - vendor/123/COPYING.123
    The following modules are missing license files:
    - LICENSE.unmatched
    - vendor/123/456/LICENSE.unmatched1
    The following license files that were specified in the configuration have changed:
    - LICENSE-Custom
    - vendor/custom/LICENSE

    GPL-3.0-only AND MIT AND Unknown

    The following license keys are NOT RECOGNIZED:
    - Unknown
    """  # noqa: E501
    assert out == dedent(expected)


def test_generate_buildrequires(capsys: pytest.CaptureFixture):
    go_vendor_license.main(["--detector=askalono", "generate_buildrequires"])
    out, err = capsys.readouterr()
    assert not err
    assert out == "askalono-cli\n"


def test_generate_buildrequires_no_check(capsys: pytest.CaptureFixture):
    go_vendor_license.main(
        ["--detector=askalono", "generate_buildrequires", "--no-check"]
    )
    out, err = capsys.readouterr()
    assert not err
    assert not out


def test_generate_buildrequires_trivy(capsys: pytest.CaptureFixture):
    go_vendor_license.main(["--detector=trivy", "generate_buildrequires"])
    out, err = capsys.readouterr()
    assert not err
    assert out == "trivy\n"


def test_generate_buildrequires_no_check_trivy(capsys: pytest.CaptureFixture):
    go_vendor_license.main(["--detector=trivy", "generate_buildrequires", "--no-check"])
    out, err = capsys.readouterr()
    assert not err
    assert out == "trivy\n"


def test_license_explicit(test_data: Path, tmp_path: Path) -> None:
    case_dir = test_data / "case1"
    licenses_dir = case_dir / "licenses"
    with open(case_dir / "config.toml", "rb") as fp:
        expected = tomllib.load(fp)
    dest = tmp_path / "config.toml"
    copy2(case_dir / "config-broken.toml", dest)
    go_vendor_license.main(
        [
            f"-c{dest}",
            f"-C{licenses_dir}",
            "explicit",
            f"-f{licenses_dir / 'LICENSE.MIT'}",
            "MIT",
        ]
    )
    with open(dest, "rb") as fp:
        gotten = tomllib.load(fp)
    assert gotten == expected



#!/usr/bin/env python3

# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Iterable

import click
from fedrq.backends.base import PackageQueryAlias, RepoqueryBase

# NOTE(gotmax23): formatters is a private API that I'm only using because I wrote it
from fedrq.cli import formatters
from fedrq.config import get_config

from go_vendor_tools.license_detection.base import LicenseDetector
from go_vendor_tools.license_detection.load import DETECTORS


def get_whatrequires_map(
    rq: RepoqueryBase,
    detectors: Iterable[type[LicenseDetector]] = DETECTORS.values(),
) -> dict[type[LicenseDetector], PackageQueryAlias]:
    result: dict[type[LicenseDetector], PackageQueryAlias] = {}
    whatbuildrequires = rq.query(requires="go-vendor-tools", arch="src")
    for detector in detectors:
        detector_wbr = rq.query(pkg=whatbuildrequires)
        for require in detector.PACKAGES_NEEDED:
            detector_wbr = detector_wbr.intersection(rq.query(requires=require))
        result[detector] = detector_wbr
    return result


@click.command(
    context_settings={"show_default": True, "help_option_names": ["-h", "--help"]}
)
@click.option("-b", "--branch")
@click.option("-r", "--repo", default="@base")
@click.option("-F", "--formatter", "formatter_name", default="plain")
def main(branch: str | None, repo: str, formatter_name: str) -> None:
    config = get_config()
    rq = config.get_rq(branch or config.default_branch, repo)
    formatter_obj = formatters.DefaultFormatters.get_formatter(
        formatter_name, repoquery=rq
    )
    wr_map = get_whatrequires_map(rq)
    for idx, (detector, detector_wbr) in enumerate(wr_map.items()):
        for require in detector.PACKAGES_NEEDED:
            detector_wbr = detector_wbr.intersection(rq.query(requires=require))
        click.secho(f"{detector.NAME} (total {len(detector_wbr)})", bold=True)
        for line in formatter_obj.format(detector_wbr):
            click.echo(line)
        if idx + 1 < len(wr_map):
            click.echo()


if __name__ == "__main__":
    main()



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Detect licenses using trivy
"""

from __future__ import annotations

import dataclasses
import json
import shutil
import subprocess
from collections.abc import Collection, Iterable, Sequence
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, TypedDict, cast

from go_vendor_tools.config.licenses import LicenseConfig
from go_vendor_tools.exceptions import LicenseError
from go_vendor_tools.license_detection.base import reuse_path_to_license_map
from go_vendor_tools.license_detection.search import (
    NOTICE_FILE_TYPE,
    find_license_files,
)
from go_vendor_tools.licensing import combine_licenses, validate_license

from .base import (
    LicenseData,
    LicenseDetector,
    LicenseDetectorNotAvailableError,
    filter_license_map,
    get_manual_license_entries,
    is_unwanted_path,
)

if TYPE_CHECKING:
    from _typeshed import StrPath


class TrivyLicenseFileEntry(TypedDict):
    Severity: str
    Category: str
    PkgName: str
    FilePath: str
    Name: str
    Confidence: float
    Link: str


class TrivyLicenseDict(TypedDict):
    Target: Literal["Loose File License(s)"]
    Class: Literal["license-file"]
    Licenses: list[TrivyLicenseFileEntry]


def run_read_json(command: Sequence[StrPath]) -> Any:
    proc: subprocess.CompletedProcess[str] = subprocess.run(
        command, check=True, text=True, capture_output=True
    )
    return json.loads(proc.stdout)


@dataclasses.dataclass(frozen=True)
class TrivyLicenseData(LicenseData):
    trivy_license_data: TrivyLicenseDict


def _load_license_data(trivy_path: StrPath, directory: StrPath) -> dict[str, Any]:
    # fmt: off
    cmd = [
        trivy_path,
        "fs",
        "--scanners", "license",
        "--license-full",
        "-f", "json",
        directory,
    ]
    # fmt: on
    return run_read_json(cmd)


def _license_data_to_trivy_license_dict(data: dict[str, Any]) -> TrivyLicenseDict:
    for item in data["Results"]:
        if item.get("Class") == "license-file":
            return cast(TrivyLicenseDict, item)
    raise ValueError("Failed to read Trivy license data")


def _trivy_license_dict_to_license_map(
    data: TrivyLicenseDict, config: LicenseConfig
) -> tuple[dict[Path, str], set[Path]]:
    license_map: dict[Path, str] = {}
    invalid: set[Path] = set()
    for result in data.get("Licenses", []):
        path = Path(result["FilePath"])
        name = result["Name"]
        if (
            # https://gitlab.com/fedora/sigs/go/go-vendor-tools/-/issues/65
            path.suffix == ".sh"
            or is_unwanted_path(
                path,
                exclude_directories=config["exclude_directories"],
                exclude_files=config["exclude_files"],
            )
        ):
            continue
        # Sometimes trivy returns names that aren't valid SPDX expressions.
        # Treat them as undetected license files in that case.
        if not validate_license(name):
            invalid.add(path)
            continue
        # License files can have multiple matches in trivy
        if path in license_map:
            license_map[path] = str(
                combine_licenses(
                    license_map[path],
                    name,
                    validate=False,
                    strict=False,
                )
            )
        else:
            license_map[path] = name
    return license_map, invalid


class TrivyLicenseDetector(LicenseDetector[TrivyLicenseData]):
    NAME = "trivy"
    PACKAGES_NEEDED = ("trivy",)
    FIND_PACKAGES_NEEDED = PACKAGES_NEEDED

    def __init__(
        self,
        detector_config: dict[str, str],
        license_config: LicenseConfig,
        find_only: bool = False,
    ) -> None:
        self._find_only = find_only
        if path := detector_config.get("trivy_path"):
            if not Path(path).exists():
                raise LicenseDetectorNotAvailableError(f"{path!r} does not exist!")
        else:
            path = shutil.which("trivy")
        if not path:
            raise LicenseDetectorNotAvailableError("Failed to find trivy binary!")

        self.path: str = path
        self.detector_config = detector_config
        self.license_config = license_config

    # TODO(anyone): Consider splitting into separate functions
    # https://gitlab.com/gotmax23/go-vendor-tools/-/issues/23
    def detect(
        self, directory: StrPath, reuse_roots: Collection[StrPath] = ()
    ) -> TrivyLicenseData:
        data = _load_license_data(self.path, directory)
        licenses = _license_data_to_trivy_license_dict(data)
        license_map, undetected = _trivy_license_dict_to_license_map(
            licenses, self.license_config
        )

        manual_license_map, manual_unmatched = get_manual_license_entries(
            self.license_config["licenses"], directory
        )
        undetected -= manual_license_map.keys()
        license_map |= manual_license_map
        license_file_lists = find_license_files(
            directory,
            relative_paths=True,
            exclude_directories=self.license_config["exclude_directories"],
            exclude_files=self.license_config["exclude_files"],
            reuse_roots=reuse_roots,
        )
        license_map |= reuse_path_to_license_map(license_file_lists["reuse"])
        license_map = dict(sorted(license_map.items(), key=lambda item: item[0]))
        # Ensure that any license files found by searching the filesystem
        # manually are detected by trivy
        undetected.update(
            file
            for file in map(Path, license_file_lists["license"])
            if file not in license_map
        )
        return TrivyLicenseData(
            directory=Path(directory),
            license_map=license_map,
            undetected_licenses=frozenset(undetected),
            unmatched_manual_licenses=manual_unmatched,
            trivy_license_data=licenses,
            extra_license_files=tuple(map(Path, license_file_lists["notice"])),
            detector_name=self.NAME,
        )

    def find_license_files(
        self, directory: StrPath, reuse_roots: Collection[StrPath] = ()
    ) -> list[Path]:
        data = _load_license_data(self.path, directory)
        licenses = _license_data_to_trivy_license_dict(data)
        license_map, undetected = _trivy_license_dict_to_license_map(
            licenses, self.license_config
        )
        filtered_license_map = filter_license_map(
            license_map,
            self.license_config["exclude_directories"],
            self.license_config["exclude_files"],
        )
        manual_license_map, unmatched = get_manual_license_entries(
            self.license_config["licenses"], directory
        )
        if unmatched:
            raise LicenseError(
                "Invalid manual license config entries:"
                + "\n"
                + "\n".join(map(str, unmatched)),
            )
        license_file_lists = find_license_files(
            directory,
            relative_paths=True,
            exclude_directories=self.license_config["exclude_directories"],
            exclude_files=self.license_config["exclude_files"],
            reuse_roots=reuse_roots,
            filetype_info=[NOTICE_FILE_TYPE],
        )
        files: set[Path] = {
            *filtered_license_map.keys(),
            *undetected,
            *manual_license_map,
            *map(Path, chain.from_iterable(license_file_lists.values())),
        }
        return sorted(files)

    def detect_files(
        self, files: Iterable[Path], directory: Path | None = None
    ) -> tuple[dict[Path, str], set[Path]]:
        raise NotImplementedError("The trivy backend does not support detect_files.")



# Copyright (C) 2025 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT
from __future__ import annotations

import enum
import fnmatch
import os
import re
from collections.abc import Collection, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from _typeshed import StrPath


# Adapted from the ignore Rust crate's patterns that askalono crawl uses
# https://github.com/BurntSushi/ripgrep/blob/79cbe89deb1151e703f4d91b19af9cdcc128b765/crates/ignore/src/default_types.rs#L123
_LICENSE_GLOBS: Sequence[str] = (
    # General
    "COPYING",
    "COPYING[.-]*",
    "COPYRIGHT",
    "COPYRIGHT[.-]*",
    "EULA",
    "EULA[.-]*",
    "licen[cs]e",
    "licen[cs]e.*",
    "LICEN[CS]E",
    "LICEN[CS]E[.-]*",
    "*[.-]LICEN[CS]E*",
    # "NOTICE", "NOTICE[.-]*",
    # "PATENTS", "PATENTS[.-]*",
    "UNLICEN[CS]E",
    "UNLICEN[CS]E[.-]*",
    # GPL (gpl.txt, etc.)
    "agpl[.-]*",
    "gpl[.-]*",
    "lgpl[.-]*",
    # Other license-specific (APACHE-2.0.txt, etc.)
    "AGPL-*[0-9]*",
    "APACHE-*[0-9]*",
    "BSD-*[0-9]*",
    "CC-BY-*",
    "GFDL-*[0-9]*",
    "GNU-*[0-9]*",
    "GPL-*[0-9]*",
    "LGPL-*[0-9]*",
    "MIT-*[0-9]*",
    "MPL-*[0-9]*",
    "OFL-*[0-9]*",
    # New patterns added by go-vendor-tools
    "Li[cs]ense",
    "Li[cs]ense.*",
    "*-license.txt",
    "*-bsd.txt",
)
# _LICENSE_PATTERNS =
_LICENSE_PATTERN = re.compile(
    "(" + "|".join(map(fnmatch.translate, _LICENSE_GLOBS)) + ")"
)

_LICENSE_EXCLUDE_PATTERN = re.compile(
    r"""
    (
        .*\.go|         # Some projects have license.go files that are code
        .*\.py|         # Some projects have license.py files that are code
        (?!)            # Dummy regex to allow a trailing "|"
    )""",
    flags=re.VERBOSE,
)
# License exclude patterns that should only apply to subdirectories of vendor/.
_LICENSE_EXCLUDE_PATTERN_SUBDIR = re.compile(
    r"""
    (
        LICENSE.docs|   # Docs from vendored libs are not included in the final package
        (?!)            # Dummy regex to allow a trailing "|"
    )""",
    flags=re.VERBOSE,
)

_NOTICE_PATTERN = re.compile(r"^(AUTHORS|NOTICE|PATENTS).*$")


class LicenseFileTypes(enum.Enum):
    LICENSE_FILE = enum.auto()
    NOTICE_FILE = enum.auto()


@dataclass(frozen=True)
class LicenseRegexFileType:
    _DISALLOWED_NAMES = ("reuse",)
    name: str
    regex: re.Pattern[str]
    exclude_regex: re.Pattern[str] | None = None
    exclude_subdir_regex: re.Pattern[str] | None = None

    def __post_init__(self) -> None:
        if self.name in self._DISALLOWED_NAMES:
            raise ValueError(f"Disallowed name: {self.name}")


LICENSE_FILE_TYPE = LicenseRegexFileType(
    "license",
    _LICENSE_PATTERN,
    _LICENSE_EXCLUDE_PATTERN,
    _LICENSE_EXCLUDE_PATTERN_SUBDIR,
)
NOTICE_FILE_TYPE = LicenseRegexFileType(
    "notice", _NOTICE_PATTERN, _LICENSE_EXCLUDE_PATTERN
)
DEFAULT_FILE_TYPES = (LICENSE_FILE_TYPE, NOTICE_FILE_TYPE)


def _clean_dirnames(dirnames: Collection[object]) -> set[str]:
    return {d.rstrip("/") for d in map(str, dirnames)}


def find_license_files(
    directory: StrPath,
    relative_paths: bool,
    exclude_directories: Collection[str] = (),
    exclude_files: Collection[str] = (),
    # FIXME(gotmax23): Properly integrate and test REUSE support
    reuse_roots: Collection[StrPath] = (),
    *,
    filetype_info: Sequence[LicenseRegexFileType] = DEFAULT_FILE_TYPES,
):
    """
    Find license files of different types

    Args:
        directory: Root directory to recurse through to find license files
        relative_paths: Whether to return paths relative to `directory` or full paths
        exclude_directories: Directory paths relative to `directory` to exclude
        exclude_files: File paths relative to `directory` to exclude
        reuse_roots: Directories to search for REUSE-style LICENSES directory
    """
    licenses: dict[str, list[str]] = {ft.name: [] for ft in filetype_info}
    licenses["reuse"] = []
    exclude_directories = _clean_dirnames(exclude_directories)
    reuse_roots = _clean_dirnames(reuse_roots)
    for root, dirnames, files in os.walk(directory):
        rootpath = os.path.relpath(root, directory)
        rootbasename = os.path.basename(root)
        rootdirname = os.path.dirname(rootpath)
        in_reuse_dir = rootbasename == "LICENSES" and (
            # "" if rootdirname is the root directory, which we always want
            # to consider as a reuse directory
            rootdirname == ""
            or rootdirname in reuse_roots
        )
        if in_reuse_dir:
            # If it's a REUSE directory, we don't need to recurse any further
            dirnames.clear()
        else:
            for dirname in list(dirnames):
                dirpath = os.path.relpath(os.path.join(root, dirname), directory)
                if dirpath in exclude_directories:
                    dirnames.remove(dirname)
        for file in files:
            fullpath = os.path.join(root, file)
            filepath = os.path.relpath(fullpath, directory)
            if filepath in exclude_files:
                continue
            if in_reuse_dir:
                licenses["reuse"].append(filepath if relative_paths else fullpath)
            else:
                for ft in filetype_info:
                    if ft.regex.fullmatch(file) and (
                        (not ft.exclude_regex or not ft.exclude_regex.fullmatch(file))
                        and (
                            # Inside the root directory so n/a
                            not rootpath.startswith(f"vendor{os.sep}")
                            # Regex is None so n/a
                            or not ft.exclude_subdir_regex
                            # Regex does not match, so okay to include
                            or not ft.exclude_subdir_regex.fullmatch(file)
                        )
                    ):
                        licenses[ft.name].append(
                            filepath if relative_paths else fullpath
                        )
                        break
    return licenses



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

import json
from pathlib import Path
from subprocess import CalledProcessError
from typing import Any

import pytest
from pytest_mock import MockerFixture

from go_vendor_tools.config.base import BaseConfig, load_config
from go_vendor_tools.license_detection.askalono import AskalonoLicenseDetector
from go_vendor_tools.license_detection.base import (
    LicenseData,
    LicenseDetector,
    get_manual_license_entries,
)
from go_vendor_tools.license_detection.load import DETECTORS
from go_vendor_tools.license_detection.trivy import TrivyLicenseDetector


def test_get_extra_licenses(test_data: Path) -> None:
    case_dir = test_data / "case1"
    licenses_dir = case_dir / "licenses"
    config = load_config(case_dir / "config.toml")
    matched, missing = get_manual_license_entries(
        config["licensing"]["licenses"], licenses_dir
    )
    expected_map = {
        Path("LICENSE.BSD3"): "BSD-3-Clause",
        Path("LICENSE.MIT"): "MIT",
    }
    assert matched == expected_map
    assert not missing


def test_get_extra_licenses_error(test_data: Path, config1_broken: BaseConfig) -> None:
    case_dir = test_data / "case1"
    licenses_dir = case_dir / "licenses"
    matched, missing = get_manual_license_entries(
        config1_broken["licensing"]["licenses"], licenses_dir
    )
    expected_map = {Path("LICENSE.BSD3"): "BSD-3-Clause"}
    assert matched == expected_map
    assert missing == (Path("LICENSE.MIT"),)


@pytest.mark.parametrize(
    "case_name, allowed_detectors, cli_config",
    [
        pytest.param("case2", None, {}, id="case2"),
        pytest.param("case3", [AskalonoLicenseDetector], {"multiple": "1"}, id="case3"),
    ],
)
def test_load_dump_license_data(
    test_data: Path,
    detector: type[LicenseDetector],
    case_name: str,
    allowed_detectors: list[type[LicenseDetector]] | None,
    cli_config: dict[str, str],
    mocker: MockerFixture,
) -> None:
    if allowed_detectors is not None and detector not in allowed_detectors:
        pytest.skip(f"{case_name} does use {detector}")

    # Needed for case3
    mocker.patch("go_vendor_tools.gomod.get_go_module_names", return_value={"abc": ""})

    case_dir = test_data / case_name
    expected_report = case_dir / "reports" / f"{detector.NAME}.json"
    licenses_dir = case_dir / "licenses"
    config = load_config(case_dir / "go-vendor-tools.toml", allow_missing=True)
    detector_obj = detector(cli_config, config["licensing"])
    try:
        data: LicenseData = detector_obj.detect(licenses_dir, ("vendor/abc",))
    except Exception as exc:
        print(exc)
        if isinstance(exc, CalledProcessError):
            print("stdout:", exc.stdout)
            print("stderr:", exc.stderr)
            if case_name == "case3" and "Found argument '--multiple'" in exc.stderr:
                # stderr: error: Found argument '--multiple' which wasn't
                # expected, or isn't valid in this context
                # For some reason, this only happens on EL 9.
                pytest.xfail()
        raise

    placeholder_path = Path("/placeholder")
    data = data.replace(directory=placeholder_path)

    jsonable = data.to_jsonable()
    new_data = type(data).from_jsonable(jsonable)
    assert new_data.to_jsonable() == jsonable

    _remove_license_scanner_data(jsonable)
    # NOTE: Uncomment this line to regenerate the test fixtures
    # (expected_report).write_text(json.dumps(jsonable, indent=2) + "\n")
    with (expected_report).open() as fp:
        gotten_json = _remove_license_scanner_data(json.load(fp))
    assert gotten_json == jsonable


def _remove_license_scanner_data(data: dict[str, Any]) -> dict[str, Any]:
    """
    Remove license-scanner specific data from license data dict, as this data
    tends to be unstable and change between versions.
    We only care about the data that's actually produced by g-v-t.
    """
    for name in DETECTORS:
        key = f"{name}_license_data"
        data.pop(key, None)
    return data


def test_detect_nothing(tmp_path: Path, detector: type[LicenseDetector]) -> None:
    """
    Ensure the code has proper error handling for when no licenses are detected
    """
    config = load_config(None)
    detector_obj = detector({}, config["licensing"])
    data: LicenseData = detector_obj.detect(tmp_path)
    assert data.directory == tmp_path
    assert not data.license_map
    assert not data.undetected_licenses
    assert not data.license_set
    assert data.license_expression is None


def test_detect_files(detector: type[LicenseDetector], test_data: Path) -> None:
    if detector is TrivyLicenseDetector:
        pytest.skip("trivy is not supported")
    config = load_config(None)
    detector_obj = detector({}, config["licensing"])
    case1 = test_data / "case1/licenses"
    files = [
        case1 / "LICENSE.BSD3",
        case1 / "LICENSE.MIT",
        test_data / "case2/licenses/LICENSE.undetected",
    ]
    files = [path.relative_to(test_data) for path in files]
    mapping, undetected = detector_obj.detect_files(files, test_data)
    expected_mapping = {
        Path("case1/licenses/LICENSE.BSD3"): "BSD-3-Clause",
        Path("case1/licenses/LICENSE.MIT"): "MIT",
    }
    expected_undetected = {Path("case2/licenses/LICENSE.undetected")}
    assert mapping == expected_mapping
    assert undetected == expected_undetected


def test_detect_files_absolute(
    detector: type[LicenseDetector], test_data: Path
) -> None:
    if detector is TrivyLicenseDetector:
        pytest.skip("trivy is not supported")
    config = load_config(None)
    detector_obj = detector({}, config["licensing"])
    case1 = test_data / "case1/licenses"
    files = [
        case1 / "LICENSE.BSD3",
        case1 / "LICENSE.MIT",
        test_data / "case2/licenses/LICENSE.undetected",
    ]
    mapping, undetected = detector_obj.detect_files(files)
    expected_mapping = {
        files[0]: "BSD-3-Clause",
        files[1]: "MIT",
    }
    expected_undetected = {files[2]}
    assert mapping == expected_mapping
    assert undetected == expected_undetected



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Base classes for handling license detection tools
"""

from __future__ import annotations

import abc
import dataclasses
import os
import re
import sys
from collections.abc import Collection, Iterable, Mapping, Sequence
from functools import partial
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Generic

from go_vendor_tools.config.licenses import LicenseConfig, LicenseEntry
from go_vendor_tools.exceptions import LicenseError
from go_vendor_tools.hashing import verify_hash
from go_vendor_tools.license_detection.search import find_license_files
from go_vendor_tools.licensing import combine_licenses, get_unknown_license_keys

if TYPE_CHECKING:
    from _typeshed import StrPath

    # TypeVar from typing_extensions needed for PEP 696
    from typing_extensions import Self, TypeVar
else:
    from typing import TypeVar

EXTRA_LICENSE_FILE_REGEX = re.compile(
    r"^(AUTHORS|NOTICE|PATENTS).*$", flags=re.IGNORECASE
)


def get_manual_license_entries(
    licenses: list[LicenseEntry], directory: StrPath
) -> tuple[dict[Path, str], tuple[Path, ...]]:
    results: dict[Path, str] = {}
    not_matched: list[Path] = []
    seen: set[Path] = set()
    for lic in licenses:
        relpath = Path(lic["path"])
        path = directory / relpath
        if path in results:
            raise LicenseError(
                f"{path} was specified multiple times in the configuration!"
            )
        seen.add(path)
        if verify_hash(path, lic["sha256sum"]):
            results[relpath] = lic["expression"]
        else:
            not_matched.append(relpath)
    return results, tuple(not_matched)


def is_unwanted_path(
    path: Path,
    exclude_directories: Collection[StrPath],
    exclude_files: Collection[str],
) -> bool:
    return (
        # Hardcoded exception
        "testdata" in path.parts
        or str(path) in exclude_files
        or any(path.is_relative_to(directory) for directory in exclude_directories)
    )


def filter_license_map(
    license_map: dict[Path, str],
    exclude_directories: Collection[str],
    exclude_files: Collection[str],
) -> dict[Path, str]:
    """
    Filter licenses files from unwanted paths
    """
    exclude_directories = set(exclude_directories)
    exclude_files = set(exclude_files)
    return {
        path: exp
        for path, exp in license_map.items()
        if not is_unwanted_path(path, exclude_directories, exclude_files)
    }


def python3dist(package: str, /) -> str:
    return f"python{sys.version_info.major}.{sys.version_info.minor}dist({package})"


# TODO(anyone): Should we check for valid filenames
# (each file should be a single license name)
def reuse_path_to_license_map(files: Collection[StrPath]) -> dict[Path, str]:
    result: dict[Path, str] = {}
    for file in files:
        name = os.path.splitext(os.path.basename(file))[0]
        result[Path(file)] = name
    return result


@dataclasses.dataclass(frozen=True)
class LicenseData:
    """
    Generic class representing detected license data.
    Can be subclassed by detector implementations to add additional fields.
    Attributes marked with (Generated fields) are calculated during class init
    and are not passed as values.
    The class is a frozen dataclass and is meant to be (pseudo) immutable.

    Attributes:
        directory:
            Path that was crawled for licensed
        license_map:
            Mapping of relative paths to license (within `directory`) to str
            SPDX license expressions
        undetected_licenses:
            Relative paths of license files that the license detector
            implementation failed to detect
        unmatched_manaul_licenses:
            Relative paths of invalid manually specified license entries
        extra_license_files:
            Relative paths to extra files (e.g., AUTHORS or NOTICE files) that
            we should include in the distribution but not run through the
            license detector
        detector_name:
            Name of the license detector
        license_set:
            (Generated field) Set of unique detected license expressions
        license_expression:
            (Generated field) Cumulative `license_expression.LicenseExpression`
            SPDX expression
        license_files_paths:
            (Generated field) Absolute paths to all detected license files
        unknown_license_keys:
            (Generated field) Unknown license keys in the license_expression.
        is_valid_license:
            (Generated field) Whether license is unknown (i.e., no
            unknown_license_keys)
    """

    directory: Path
    license_map: Mapping[Path, str]
    undetected_licenses: frozenset[Path]
    unmatched_manual_licenses: tuple[Path, ...]
    extra_license_files: tuple[Path, ...]
    detector_name: str

    # Generated fields
    license_set: frozenset[str] = dataclasses.field(init=False, compare=False)
    license_expression: str = dataclasses.field(init=False, compare=False)
    license_file_paths: tuple[Path, ...] = dataclasses.field(init=False, compare=False)
    unknown_license_keys: list[str] = dataclasses.field(init=False, compare=False)
    is_valid_license: bool = dataclasses.field(init=False, compare=False)

    # Helper for *_jsonable() methods
    _LIST_PATH_FIELDS: ClassVar = (
        "undetected_licenses",
        "unmatched_manual_licenses",
        "license_file_paths",
        "extra_license_files",
    )

    # Implement replace() method
    replace = dataclasses.replace

    def __post_init__(self) -> None:
        # Use object.__setattr__ because this is a frozen dataclass.
        object.__setattr__(self, "license_set", frozenset(self.license_map.values()))
        object.__setattr__(
            self,
            "license_expression",
            self._combine_licenses(*self.license_set) if self.license_map else None,
        )
        object.__setattr__(
            self,
            "license_file_paths",
            tuple(
                sorted(
                    self.directory / lic
                    for lic in chain(self.license_map, self.undetected_licenses)
                ),
            ),
        )
        object.__setattr__(
            self,
            "unknown_license_keys",
            get_unknown_license_keys(self.license_expression),
        )
        object.__setattr__(self, "is_valid_license", not self.unknown_license_keys)

    _combine_licenses = staticmethod(
        partial(combine_licenses, validate=False, strict=False)
    )

    # This would be a good task for pydantic, but we want to keep dependencies slim.
    def to_jsonable(self) -> dict[str, Any]:
        data = dataclasses.asdict(self)
        for key, value in data.items():
            if key == "directory":
                data[key] = str(value)
            elif key == "license_map":
                data[key] = {str(key1): value1 for key1, value1 in value.items()}
            elif key in self._LIST_PATH_FIELDS:
                data[key] = list(map(str, value))
                if not isinstance(value, Sequence):
                    data[key].sort()
            elif key == "license_set":
                data[key] = sorted(value)
            elif key == "license_expression":
                data[key] = str(value)
        return data

    @classmethod
    def _from_jsonable_to_dict(cls, data: dict[Any, Any]) -> dict[Any, Any]:
        init_fields = [field.name for field in dataclasses.fields(cls) if field.init]
        newdata: dict[Any, Any] = {}
        for key, value in data.items():
            if key not in init_fields:
                continue
            if key == "directory":
                newdata[key] = Path(value)
            elif key == "license_map":
                newdata[key] = {Path(key1): value1 for key1, value1 in value.items()}
            elif key in cls._LIST_PATH_FIELDS:
                func = set if key == "undetected_licenses" else tuple
                newdata[key] = func(map(Path, value))
            else:
                newdata[key] = value
        return newdata

    @classmethod
    def from_jsonable(cls, data: dict[Any, Any]) -> Self:
        return cls(**cls._from_jsonable_to_dict(data))


if TYPE_CHECKING:
    _LicenseDataT_co = TypeVar(
        "_LicenseDataT_co", bound=LicenseData, covariant=True, default=LicenseData
    )
else:
    _LicenseDataT_co = TypeVar("_LicenseDataT_co", covariant=True, bound=LicenseData)


class LicenseDetector(Generic[_LicenseDataT_co], metaclass=abc.ABCMeta):
    """
    ABC for a license detector backend

    Attributes:
        NAME: Name of the license detector
        PACKAGES_NEEDED:
            Tuple of Fedora package names needed for the license detector
        FIND_PACKAGES_NEEDED:
            Tuple of packages needed for find_only mode (see __init__ docstring)
        license_config:
            LicenseConfig object passed to the constructor
        detector_config:
            Options passeed to constructor
        find_only: Whether find_only mode is enabled
    """

    NAME: ClassVar[str]
    PACKAGES_NEEDED: ClassVar[tuple[str, ...]] = ()
    FIND_PACKAGES_NEEDED: ClassVar[tuple[str, ...]] = ()
    detector_config: dict[str, str]
    license_config: LicenseConfig
    _find_only: bool

    @abc.abstractmethod
    def __init__(
        self,
        detector_config: dict[str, str],
        license_config: LicenseConfig,
        find_only: bool = False,
    ) -> None:
        """
        Args:
            detector_config:
                String key-value pairs of --detector-config options that are
                defined separately for each license detector implementation
            license_config:
                LicenseConfig object.
                The detector_config option is ignored in favor of the
                detector_config argument.
            find_only:
                When find_only is enabled, only the dependencies for the
                find_license_files method is checked.
                This allows a lightweight mode without the dependencies for the
                detect() method when only a list of valid license files is
                required.
        """

    @property
    def find_only(self):
        """
        Whether find_only mode is enabled.
        """
        return self._find_only

    @abc.abstractmethod
    def detect(
        self, directory: StrPath, reuse_roots: Collection[StrPath] = ...
    ) -> _LicenseDataT_co:
        """
        Scan a directory for license data

        Args:
            directory: Directory
            reuse_roots: Directories to search for REUSE-style LICENSES directory
        """

    @abc.abstractmethod
    def detect_files(
        self, files: Iterable[Path], directory: Path | None = None
    ) -> tuple[dict[Path, str], set[Path]]:
        """
        Given a list of license files, return a mapping of paths to license expressions.

        Args:
            files: List files
            directory:
                Directory to which paths are relative or None to treat as
                absolute paths.

        Returns: (License mapping, undetected files)
        """

    def find_license_files(
        self, directory: StrPath, reuse_roots: Collection[StrPath] = ()
    ) -> list[Path]:
        """
        Default implementation of find_license_files.

        Args:
            directory: Directory
            reuse_roots: Directories to search for REUSE-style LICENSES directory
        Raises:
            LicenseError:
                Invalid manual license config entries are present in the license config
        """
        license_file_lists = find_license_files(
            directory,
            relative_paths=True,
            exclude_directories=self.license_config["exclude_directories"],
            exclude_files=self.license_config["exclude_files"],
            reuse_roots=reuse_roots,
        )
        manual_license_map, unmatched = get_manual_license_entries(
            self.license_config["licenses"], directory
        )
        if unmatched:
            raise LicenseError(
                "Invalid manual license config entries:"
                + "\n"
                + "\n".join(map(str, unmatched)),
            )
        files: set[Path] = {
            Path(p) for p in chain.from_iterable(license_file_lists.values())
        }
        files.update(manual_license_map)
        return sorted(files)


class LicenseDetectorNotAvailableError(LicenseError):
    """
    Failed to load the requested license detector
    """



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Parse go.mod and modules.txt files
"""

from __future__ import annotations

import re
from collections.abc import Collection
from pathlib import Path

MODULE_REGEX = re.compile(r"^# (?:.+=>)?(?P<ipath>\S+) v(?P<version>\S+)$")


def get_go_module_names(directory: Path, allow_missing: bool = True) -> dict[str, str]:
    results: dict[str, str] = {}
    try:
        fp = (directory / "vendor/modules.txt").open("r", encoding="utf-8")
    except FileNotFoundError:
        if not allow_missing:
            raise
        return results
    with fp:
        for line in fp:
            if match := MODULE_REGEX.match(line):
                results[match["ipath"]] = match["version"]
    return results


# TODO: Test go_mod_dirs support
def get_go_module_dirs(
    directory: Path,
    relative_paths: bool = False,
    go_mod_dir: str | None = None,
    go_module_names: Collection[str] | None = None,
) -> list[Path]:
    go_mod_dir = go_mod_dir or "."
    results: list[Path] = []
    for ipath in (
        go_module_names
        if go_module_names is not None
        else get_go_module_names(directory)
    ):
        moddir = directory / go_mod_dir / "vendor" / ipath
        if moddir.is_dir():
            results.append(
                moddir.relative_to(directory) if relative_paths else moddir.resolve()
            )
    return results


def get_unlicensed_mods(
    directory: Path,
    license_paths: Collection[Path],
    go_mod_dir: str | None = None,
    go_module_names: Collection[str] | None = None,
) -> set[Path]:
    resolved_dir = directory.resolve()
    licensed_dirs = {
        (
            first.parent
            if (first := path.parent).relative_to(resolved_dir).name == "LICENSES"
            else first
        )
        for path in (p.resolve() for p in license_paths)
    }
    all_dirs = {
        *get_go_module_dirs(
            directory, go_mod_dir=go_mod_dir, go_module_names=go_module_names
        ),
        directory.resolve(),
    }
    return all_dirs - licensed_dirs



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT
# Based on <https://git.sr.ht/~gotmax23/zstarfile/tree/main/item/src/zstarfile/extra.py>.

"""
Extra functions for dealing with archives that are not part of tarfile.
Vendored from zstarfile.
"""

from __future__ import annotations

from collections.abc import Callable
from inspect import signature
from pathlib import Path
from tarfile import TarFile
from typing import TYPE_CHECKING

from go_vendor_tools.archive import OurTarFile

if TYPE_CHECKING:
    from _typeshed import StrPath


def _get_opener(
    filename: str,
    compression_type: str | None = None,
    tarfile_class: type[TarFile] = OurTarFile,
) -> Callable[..., TarFile]:
    """
    Return a function that can be used to open `filename`.
    """
    open_meths = dict(tarfile_class.OPEN_METH)
    method: str | None = None
    tar_method = open_meths.pop("tar")
    if compression_type:
        if compression_type not in open_meths:
            raise ValueError(f"Invalid compression_type: {compression_type}")
        method = open_meths[compression_type]
    elif filename.endswith(".tar"):
        method = tar_method
    else:
        for ext, meth in open_meths.items():
            if filename.endswith((f".t{ext}", f".tar.{ext}")):
                method = meth
                break
    if not method:
        raise ValueError(f"No match found for {filename}")
    return getattr(tarfile_class, method)


def open_write_compressed(
    file: StrPath,
    /,
    compression_type: str | None = None,
    compresslevel: int | None = None,
    tarfile_class: type[TarFile] = OurTarFile,
    **kwargs,
) -> TarFile:
    """
    Open a TarFile for writing and automatically detect the compression type
    based on the filename

    Args:
        file: File path
        compression_type:
            Compression type, such as `tar` (uncompressed), `gz`, or `bz2`.
            By default, detect filetype based on the filename
        compresslevel:
            An integer compression level.
            If `compresslevel` is None, do not pass the kwarg down to pass the
            kwarg down to the TarFile constructor.
        tarfile_class:
            [`tarfile.TarFile`][tarfile.TarFile] or a subclass
    """
    kwargs.pop("mode", None)
    opener = _get_opener(Path(file).name, compression_type, tarfile_class)
    sig_params = signature(opener).parameters
    if compresslevel is not None:
        level_param_name: str | None = None
        if "compresslevel" in sig_params:
            level_param_name = "compresslevel"
        elif "level" in sig_params:
            level_param_name = "level"
        if level_param_name is None:
            raise ValueError(
                f"compresslevel is not a valid option for {opener.__name__}"
            )
        kwargs[level_param_name] = compresslevel
    return opener(file, "w", **kwargs)


__all__ = ("open_write_compressed",)



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Helpers for working with specfiles
"""

from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING, Protocol, TypeVar, cast

# specfile is opened liked this so downstreams can remove the dependency on
# specfile and use the other functionality if they so choose
try:
    import specfile
    import specfile.sections
    import specfile.sources
    import specfile.tags
    from specfile.exceptions import SpecfileException
except ImportError:
    HAS_SPECFILE = False
else:
    HAS_SPECFILE = True

from go_vendor_tools.exceptions import VendorToolsError

_T_co = TypeVar("_T_co", covariant=True)

if TYPE_CHECKING:
    from contextlib import AbstractContextManager

    from typing_extensions import Self

    class _ContentContext(AbstractContextManager[_T_co], Protocol[_T_co]):
        @property
        def content(self) -> _T_co: ...


def _get_section_name(subpackage: str | None = None) -> str:
    """
    Parse a subpackage name into a %package call that can be passed to
    Sections.get()
    """
    section = "package"
    if subpackage is None:
        pass
    elif subpackage.startswith("-"):
        section += f" {subpackage[1:]}"
    else:
        section += f" -n {subpackage}"
    return section


class VendorSpecfile:
    """
    Wrapper around specfile.Specfile for working with vendor specfiles.
    Must be created using the context manager protocol.
    """

    def __init__(
        self,
        spec_path: Path | str,
        subpackage_name: str | None = None,
    ) -> None:
        """
        Args:
            spec_path: Path to specfile
            subpackage_name:
                Name of a subpackage.
                - "-NAME" for "%package NAME"
                - "NAME" for "%package -n NAME"
                - None for the main package
        """
        if not HAS_SPECFILE:
            raise ImportError("This functionality requires python3-specfile!")

        # For testing purposes only!
        extra_macros: list[tuple[str, str | None]] = []
        if os.environ.get("__GVT_TEST_MACROS"):
            for macro in ("go_vendor_license_install", "go_vendor_license_check"):
                extra_macros.append((macro, ""))  # noqa: PERF401

        try:
            self._spec = specfile.Specfile(
                spec_path, autosave=True, macros=extra_macros
            )
        except SpecfileException as exc:
            raise VendorToolsError(
                f"Failed to parse specfile {spec_path}: {exc}"
            ) from exc
        self.subpackage_name = subpackage_name

    # Interact with specfile

    @property
    def license(self) -> str:
        se = VendorToolsError(
            "Failed to access License tag in "
            + _get_section_name(self.subpackage_name),
        )
        try:
            value = self.tags().content.get("License").expanded_value
        except AttributeError as exc:
            raise se from exc
        if value is None:
            raise se
        return value

    @license.setter
    def license(self, value: str) -> None:
        with self.tags() as tags:
            try:
                tags.get("License").value = value
            except AttributeError as exc:
                raise VendorToolsError(
                    "Failed to set License tag in "
                    + _get_section_name(self.subpackage_name)
                ) from exc

    def source0_and_source1(self, directory: Path | None = None) -> tuple[Path, Path]:
        """
        Assuming that Source0 is the main archive and Source1 is the vendor archive,
        return absolute paths to each.

        Args:
            directory: Sources directory (defaults to spec path's dirname)

        Yields: (Source0 path, Source1 path)

        Raises:
            VendorToolsError: If Source0 and/or Source1 don't exist
        """
        spec_path = cast(Path, self.spec.path)
        directory = directory or spec_path.resolve().parent
        se = VendorToolsError(f"Source0 and Source1 must be specified in {spec_path}")
        try:
            source0, source1 = self.sources().content[:2]
            source0_name = source0.expanded_filename
            source1_name = source1.expanded_filename
        except IndexError:
            raise se from None
        if not source0_name or not source1_name:
            raise se
        return (directory / source0_name, directory / source1_name)

    # Lower-level
    @property
    def spec(self) -> specfile.Specfile:
        return self._spec

    def sources(self) -> _ContentContext[specfile.sources.Sources]:
        """
        Return a specfile Sources object associated with self.spec.
        """
        return self.spec.sources()

    def tags(
        self, subpackage_name: str | None = None
    ) -> _ContentContext[specfile.tags.Tags]:
        """
        Return a Tags context manager associated with self.subpackage_name

        Args:
            subpackage_name: Defaults to self.subpackage_name
        """

        return self.spec.tags(
            _get_section_name(subpackage_name or self.subpackage_name)
        )

    def save(self) -> None:
        """
        Save specfile changes to disk
        """
        self.spec.save()

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *_) -> None:
        self.save()



# Copyright (C) 2025 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

import argparse
import sys
from textwrap import dedent, indent


def format_choices(*choices: object):
    result = ""
    length = len(choices)
    for idx, choice in enumerate(choices):
        result += f"`{choice}`"
        if idx < length - 1:
            result += ", "
    return result


def head_factory(initial_indent_level: int):
    return lambda level=1: "#" * (initial_indent_level + level)


def format_subparsers(
    parser: argparse.ArgumentParser, initial_indent_level: int
) -> list[str]:
    head = head_factory(initial_indent_level)
    lines: list[str] = []
    subparsers_action: argparse._SubParsersAction
    if not parser._subparsers:
        return lines
    for action in reversed(parser._subparsers._actions):
        if isinstance(action, argparse._SubParsersAction):
            subparsers_action = action
            break
    else:
        return lines
    lines.append(f"\n{head(1)} Subcommands")
    for subparser in subparsers_action.choices.values():
        if "internal" in (subparser.description or "").lower():
            continue
        lines.append("")
        lines.extend(format_command(subparser, initial_indent_level + 1))
    return lines


def _process_actions(
    actions: list[argparse.Action],
) -> tuple[list[argparse.Action], list[argparse.Action]]:
    """
    Returns: (regular actions, positional arg actions)
    """
    options: list[argparse.Action] = []
    positionals: list[argparse.Action] = []
    for action in actions:
        if isinstance(action, argparse._SubParsersAction):
            continue
        if action.option_strings:
            options.append(action)
        else:
            assert action.dest
            positionals.append(action)
    return options, positionals


def _format_action(action: argparse.Action) -> list[str]:
    lines: list[str] = []
    if action.option_strings:
        # Option
        opts = [f"`{opt}`" for opt in action.option_strings]
        line = "- " + ",".join(opts)
        if action.metavar:
            line += f" `{action.metavar}`"
    else:
        metavar = action.metavar or action.dest
        line = f"- `{metavar}`"
    if action.help:
        action_vars = vars(action).copy()
        if action_vars.get("choices"):
            action_vars["choices"] = format_choices(*action_vars["choices"])
        help_msg = dedent(action.help)
        help_msg = help_msg % action_vars
        if help_msg.startswith("\n"):
            lines.append(line + ":")
            line = indent(help_msg, " " * 4)
        else:
            line += f": {help_msg}"
    lines.append(line)
    return lines


def _format_actions(actions: list[argparse.Action]) -> list[str]:
    lines: list[str] = []
    for action in actions:
        lines.extend(_format_action(action))
    return lines


def format_all_actions(
    actions: list[argparse.Action], initial_indent_level: int
) -> list[str]:
    head = head_factory(initial_indent_level)
    lines: list[str] = []
    options, positionals = _process_actions(actions)
    if positionals:
        lines.append(f"\n{head()} Positionals\n")
        lines.extend(_format_actions(positionals))
    if options:
        lines.append(f"\n{head()} Options\n")
        lines.extend(_format_actions(options))
    return lines


def format_command(
    parser: argparse.ArgumentParser,
    initial_indent_level: int = 0,
) -> list[str]:
    head = head_factory(initial_indent_level)

    lines = []
    lines.append(f"{head()} {parser.prog}")
    if parser.description:
        lines.append(f"\n{head(2)} Description\n")
        lines.append(parser.description)
    lines.append(f"\n{head(2)} Usage\n")
    lines.append("``` text")
    lines.append(parser.format_usage().strip("\n").removeprefix("usage: "))
    lines.append("```")
    lines.extend(format_all_actions(parser._actions, initial_indent_level + 1))
    lines.extend(format_subparsers(parser, initial_indent_level + 1))
    return lines


def get_lines(refresh: bool = False) -> list[str]:
    """
    Args:
        refresh:
            When this is run with mkdocs serve, it is necessary to reload the
            module to ensure the latest version of get_parser is used.
    """
    if refresh:
        sys.modules.pop("go_vendor_tools.cli.go_vendor_license", None)
    from go_vendor_tools.cli.go_vendor_license import get_parser  # noqa: PLC0415

    parser = get_parser()
    return format_command(parser)


def main() -> None:
    print("\n".join(get_lines()))


if __name__ == "__main__":
    main()



# Copyright (C) 2025 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT
from __future__ import annotations

import sys
from collections.abc import Generator, Iterable
from enum import Enum, auto
from pathlib import Path
from typing import TYPE_CHECKING, Any

from pytest import MonkeyPatch

from go_vendor_tools.cli import go_vendor_license
from go_vendor_tools.license_detection.base import LicenseData

if TYPE_CHECKING:
    from typing_extensions import Self, TypeAlias


class CallAction(Enum):
    PRINT = auto()
    INPUT = auto()


# TODO(gotmax23): WHY DOES THIS NOT WORK ON EPEL 9 BUT ONLY SOMETIMES?
# git blame this comment for the error message.
# @dataclasses.dataclass
class Call:  # noqa: PLW1641
    action: CallAction
    args: tuple[object, ...]
    kwargs: dict[str, Any]

    # Implement the methods manually instead...
    def __init__(
        self, action: CallAction, args: tuple[object, ...], kwargs: dict[str, Any]
    ) -> None:
        self.action = action
        self.args = args
        self.kwargs = kwargs

    def __repr__(self) -> str:
        action = self.action
        args = self.args
        kwargs = self.kwargs
        return f"{type(self).__name__}({action=}, {args=}, {kwargs=})"

    def __eq__(self, other: object, /) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return (
            self.action == other.action
            and self.args == other.args
            and self.kwargs == other.kwargs
        )

    @classmethod
    def print(cls, *args, **kwargs) -> Self:
        return cls(CallAction.PRINT, args, kwargs)

    @classmethod
    def input(cls, *args, **kwargs) -> Self:
        return cls(CallAction.INPUT, args, kwargs)


TestGenType: TypeAlias = "Generator[Any, Call, None]"


class AssertCall:
    def __init__(self) -> None:
        self.nassert = 0

    def assert_call(self, call1: Call, call2: Call) -> None:
        assert call1 == call2, f"assert {self.nassert} failed"
        self.nassert += 1


# TODO(anyone): Test combinations.
# - Both disabled
# - Nothing to fill
# - Prompt w/o autofill
# - Autofill w/o prompt
# - EXCLUDE
def generator(directory: Path) -> TestGenType:
    assert_call = AssertCall().assert_call
    p = Call.print
    i = Call.input
    yield

    assert_call(
        (yield),
        p(
            "The test_autofill backend will be used to autofill missing licenses",
            file=sys.stderr,
        ),
    )
    yield None

    assert_call((yield), p("Autofilled 1 manual license entries", file=sys.stderr))
    yield None

    assert_call((yield), p("Undetected licenses found! Please enter them manually."))
    yield None

    assert_call((yield), p(f"* Undetected license: {directory / 'LICENSE.MIT'}"))
    yield None

    assert_call((yield), i("Enter SPDX expression (or EXCLUDE): "))
    yield "MIT"

    assert_call((yield), p("Expression simplified to 'MIT'"))
    yield None


class Helper:
    def __init__(self, gen: TestGenType) -> None:
        self._gen = gen
        next(gen)

    def send(self, call: Call) -> Any:
        try:
            next(self._gen)
            return self._gen.send(call)
        except StopIteration:
            assert not call, "The generator ran out"
            raise

    def print(self, *args, **kwargs):
        return self.send(Call.print(*args, **kwargs))

    def input(self, *args, **kwargs):
        return self.send(Call.input(*args, **kwargs))

    def finish(self) -> None:
        try:
            next(self._gen)
        except StopIteration:
            return
        else:
            raise AssertionError("Generator did not finish as expected")


class TestAutofill:
    NAME = "test_autofill"

    def detect_files(
        self, files: Iterable[Path], directory: Path | None = None
    ) -> tuple[dict[Path, str], set[Path]]:
        files = set(files)
        assert files == {Path("LICENSE.GPL3"), Path("LICENSE.MIT")}
        files.remove(Path("LICENSE.GPL3"))
        return {Path("LICENSE.GPL3"): "GPL-3.0-or-later"}, files


def test_get_missing_licenses(monkeypatch: MonkeyPatch, test_data: Path) -> None:
    directory = test_data / "test_get_missing_licenses"
    helper = Helper(generator(directory))
    monkeypatch.setattr(go_vendor_license, "print", helper.print, False)
    monkeypatch.setattr(go_vendor_license, "input", helper.input, False)
    data = LicenseData(
        directory=directory,
        license_map={Path("LICENSE.BSD3"): "BSD-3-Clause"},
        undetected_licenses=frozenset({Path("LICENSE.GPL3"), Path("LICENSE.MIT")}),
        unmatched_manual_licenses=(),
        extra_license_files=(),
        detector_name="test",
    )
    expected_data = LicenseData(
        directory=directory,
        license_map={
            Path("LICENSE.BSD3"): "BSD-3-Clause",
            Path("LICENSE.GPL3"): "GPL-3.0-or-later",
            Path("LICENSE.MIT"): "MIT",
        },
        undetected_licenses=frozenset(),
        unmatched_manual_licenses=(),
        extra_license_files=(),
        detector_name="test",
    )
    config: Any = {"licensing": {"licenses": []}}
    expected_config: Any = {
        "licensing": {
            "licenses": [
                {
                    "expression": "GPL-3.0-or-later",
                    "path": "LICENSE.GPL3",
                    "sha256sum": "13fa41e00e7fea297e91099d24670d0f237a7ea853ba370330b2ea0a7fe2dcfb",  # noqa: E501
                },
                {
                    "expression": "MIT",
                    "path": "LICENSE.MIT",
                    "sha256sum": "111fc6ef9cb7a4932b0a11af2e0323b2a2130233dc1f4fa11b13cff74afc4cd9",  # noqa: E501
                },
            ]
        }
    }
    test_autofill = TestAutofill()
    gotten_data = go_vendor_license.fill_missing_licenses(
        data, config, test_autofill, True  # type: ignore
    )
    helper.finish()
    assert gotten_data == expected_data
    assert config == expected_config



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
# Copyright (C) 2025 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

# TODO: Once this code stablizies, we should consider moving it into
# go-rpm-macros.

# TODO: Add unit tests. This is currently 92% covered, but only by integration
# tests.


"""
Rewritten gocheck2 that avoids failing fast and works with GO111MODULE enabled.
This is meant to be run using the %gocheck2 macro and not directly.
"""

from __future__ import annotations

import argparse
import dataclasses
import json
import os
import shlex
import subprocess
import sys
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Any

try:
    import argcomplete
except ImportError:
    HAS_ARGCOMPLETE = False
else:
    HAS_ARGCOMPLETE = True


# Using a partial breaks pytest capsys
def eprint(*args: object, **kwargs: Any) -> None:
    kwargs.setdefault("file", sys.stderr)
    kwargs.setdefault("flush", True)
    return print(*args, **kwargs)


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description=__doc__, prog="%gocheck2")
    # NOTE: See the comment about determining the primary goipath in the main function.
    # Allowing to pass custom paths complicates this.
    # Therefore, we might consider removing this argument before the final release.
    parser.add_argument(
        "-p",
        "--path",
        action="append",
        help="Relative paths that include go.mod packages."
        " This flag is SUBJECT TO CHANGE."
        " It is recommended to cd to the correct directory before running"
        " %%gocheck2 and not use this option.",
    )
    parser.add_argument(
        "-F",
        "--no-follow",
        help="Don't search for Go submodules (i.e., go.mod files in subdirectories)",
        action="store_false",
        dest="follow",
    )
    parser.add_argument(
        "-s",
        "--skip",
        action="append",
        help="Skip individual test function names. Can be repeated.",
    )
    parser.add_argument(
        "-d",
        "--directory",
        action="append",
        help="Exclude the files contained in DIRECTORY non-recursively."
        " This accepts either an import path or a path relative to"
        " the go import path of the go.mod in the current directory.",
    )
    parser.add_argument(
        "-t",
        "--tree",
        action="append",
        help="Exclude the files contained in DIRECTORY recursively."
        " This accepts either an import path or a path relative to"
        " the go import path of the go.mod in the current directory.",
    )
    parser.add_argument(
        "-L",
        "--list",
        help="List import paths to test, but don't run anything.",
        action="store_true",
    )
    parser.add_argument("extra_args", nargs="*")
    if HAS_ARGCOMPLETE:
        # fmt: off
        argcomplete.autocomplete(parser) # pyright: ignore[reportPossiblyUnboundVariable]
        # fmt: on
    return parser


def normpath_set(data: Iterable[str] | None) -> set[str]:
    return {os.path.normpath(p) for p in data} if data else set()


def parseargs() -> Args:
    ns = get_parser().parse_args()
    if ns.path:
        for path in ns.path:
            if os.path.isabs(path):
                sys.exit(f"Invalid absolute path: {path!r}. Paths must be relative!")
    return Args(
        paths=ns.path or ["."],
        ignore_dirs=normpath_set(ns.directory),
        ignore_trees=normpath_set(ns.tree),
        list_only=ns.list,
        extra_args=ns.extra_args or [],
        follow=ns.follow,
        test_skips=ns.skip or [],
    )


@dataclasses.dataclass
class Args:
    """
    Attributes:
        ignore_dirs: See -d in the argparser
        ignore_trees: See -t in the argparser
        list_only: See -L in the argparser
        test_skips: See -s in the argparser
        extra_args: Extra arguments to pass to go test.
        follow: Whether to search for go.mod in subdirectories and run tests there
        paths: Directories to find go.mod file in
    """

    ignore_dirs: set[str]
    ignore_trees: set[str]
    list_only: bool
    test_skips: list[str]
    extra_args: list[str]
    follow: bool
    paths: list[str]
    test_paths_seen: set[str] = dataclasses.field(init=False, default_factory=set)


@dataclasses.dataclass(frozen=True)
class GoModResult:
    gomod: str
    directory: str
    goipath: str


def logrun(cmd: Sequence[object], cwd: str = ".") -> None:
    r = "$ " if cwd == "." else f"({cwd}) $ "
    r += shlex.join(map(str, cmd))
    eprint(r)


def dir_okay(args: Args, path: str, goipath: str | None = None) -> bool:
    """
    Check if a directory is ignored.

    Returns: True if okay to test and false if ignored
    """
    path = path.removeprefix("./")
    paths = {path}
    if goipath and (stripped := path.removeprefix(goipath + "/")) != path:
        paths.add(stripped)
    if paths & args.ignore_dirs:
        return False
    for p in paths:
        p += "/"
        for tree in args.ignore_trees:
            if p.startswith(tree + "/"):
                return False
    return True


# args.follow is True
def _find_go_mods_follow(args: Args) -> list[GoModResult]:
    gomods: list[GoModResult] = []
    for path in args.paths:
        for root, dirnames, files in os.walk(path):
            for dirname in list(dirnames):
                dirpath = os.path.join(root, dirname)
                if dirpath.removeprefix("./") in args.ignore_trees:
                    dirnames.remove(dirname)
            if root.removeprefix("./") in args.ignore_dirs:
                continue
            for file in files:
                if file == "go.mod":
                    gomod = os.path.join(root, file)
                    goipath = get_goipath(gomod)
                    gomods.append(
                        GoModResult(gomod=gomod, directory=root, goipath=goipath)
                    )
                    break
            # https://book.pythontips.com/en/latest/for_-_else.html
            else:
                if root != str(path):
                    continue
                if path == ".":
                    msg = "No go.mod found in current directory"
                else:
                    msg = f"No go.mod found in {path}"
                sys.exit(msg)
    return gomods


# args.follow is False
def _find_go_mods_nofollow(args: Args) -> list[GoModResult]:
    gomods: list[GoModResult] = []
    for path in args.paths:
        gomod = os.path.join(path, "go.mod")
        if not os.path.isfile(gomod):
            sys.exit(f"{gomod!r} does not exist!")
        gomods.append(
            GoModResult(gomod=gomod, directory=path, goipath=get_goipath(gomod))
        )
    return gomods


def find_go_mods(args: Args) -> list[GoModResult]:
    return _find_go_mods_follow(args) if args.follow else _find_go_mods_nofollow(args)


def get_goipath(gomod: str | Path = "go.mod") -> str:
    cmd: list[str] = ["go", "mod", "edit", "-json", str(gomod)]
    logrun(cmd)
    proc = subprocess.run(cmd, stdout=subprocess.PIPE, text=True, check=False)
    data = json.loads(proc.stdout) if proc.returncode == 0 else {}
    try:
        goipath = data["Module"]["Path"]
    except KeyError:
        raise ValueError(f"Failed to retrieve Go import path from {gomod}") from None
    return goipath


def list_test_packages(
    args: Args, paths: Iterable[str], cwd: str = ".", primary_goipath: str | None = None
):
    # go list command is based on one in kubernetes Makefile.
    # The command is considered not copyrightable so the kubernetes license
    # does not apply.
    cmd = [
        "go",
        "list",
        "-find",
        "-f",
        "{{if or (gt (len .TestGoFiles) 0) (gt (len .XTestGoFiles) 0)}}{{.ImportPath}}{{end}}",  # noqa: E501,
    ]
    if tags := os.environ.get("GO_BUILDTAGS"):
        cmd.extend(("-tags", tags))
    cmd.extend(f"{goipath}/..." for goipath in paths)
    del paths  # Used up iterable
    logrun(cmd, cwd)
    proc = subprocess.run(
        cmd,
        check=True,
        text=True,
        stdout=subprocess.PIPE,
        cwd=cwd,
    )
    gotten_goipaths = proc.stdout.splitlines()
    result: list[str] = []
    for goipath in gotten_goipaths:
        if goipath in args.test_paths_seen:
            continue
        args.test_paths_seen.add(goipath)
        if dir_okay(args, goipath, primary_goipath):
            result.append(goipath)
    return result


def dogomod(args: Args, gomod: GoModResult, primary_goipath: str | None) -> int:
    test_packages = list_test_packages(
        args, [gomod.goipath], gomod.directory, primary_goipath
    )
    print(f"# {gomod.gomod}: {gomod.goipath}")
    if not test_packages:
        eprint(
            f"No test packages found for {gomod.goipath} in directory {gomod.directory}"
        )
        return 0
    if args.list_only:
        print("\n".join(test_packages))
        return 0
    extra_args = args.extra_args
    if tags := os.environ.get("GO_BUILDTAGS"):
        extra_args.extend(("-tags", tags))
    if gotest_flags := os.environ.get("GOCHECK2_GOTEST_FLAGS"):
        extra_args.extend(shlex.split(gotest_flags))
    if args.test_skips:
        extra_args.extend(("-skip", "|".join(args.test_skips)))
    cmd = ["go", "test", *extra_args, *test_packages]
    logrun(cmd, gomod.directory)
    proc = subprocess.run(cmd, check=False, cwd=gomod.directory)
    if proc.returncode != 0:
        eprint(f"Command failed with rc {proc.returncode}!")
    return proc.returncode


def main() -> None:
    if os.environ.get("GO111MODULE") == "off":
        sys.exit("G0111MODULE=off is not supported by %gocheck2. Use %gocheck instead!")
    args = parseargs()
    go_mods = find_go_mods(args)
    if not go_mods:
        sys.exit("No go.mod files found!")
    # We need to determine the root directory's go module so we can apply
    # ignore patterns that reference relative paths like the old %gocheck did.
    # This is imperfect, because with Go modules, the directory structure does
    # not have to match go module paths.
    # For example, in the etcd project:
    #  go.mod in the root directory has "go.etcd.io/etcd/v3" but api/go.mod has
    #  "go.etcd.io/etcd/api/v3," not "go.etcd.io/etcd/v3/api."
    # This means that passing "-t api" will not work.
    # Instead, it's necessary to use the full path (-t go.etcd.io/etcd/v3/api).
    if go_mods[0].directory == ".":
        primary_goipath = go_mods[0].goipath
    elif os.path.exists("go.mod"):
        primary_goipath = get_goipath()
    else:
        # TODO: Should this be made an error? Should we require the directory
        # gocheck2 is run from to have a go.mod instead of allowing it to be in
        # a subdirectory?
        eprint(
            "WARNING: No go.mod file found in CWD."
            " -d and -t flags may not work properly."
        )
        primary_goipath = None
    # TODO: Do something (e.g., show pass/fail stats by Go modules) with the
    # gomod dict keys or just make this into a simple list of dogomod() return
    # values.
    results = {gomod: dogomod(args, gomod, primary_goipath) for gomod in go_mods}
    sys.exit(max(results.values()))


if __name__ == "__main__":
    main()



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

from pathlib import Path

import pytest

from go_vendor_tools.config.base import BaseConfig, load_config
from go_vendor_tools.license_detection.base import LicenseDetector
from go_vendor_tools.license_detection.load import get_detectors as gd

HERE = Path(__file__).resolve().parent
TEST_DATA = HERE / "test_data"
CONFIG1 = load_config(TEST_DATA / "case1" / "config.toml")


@pytest.fixture
def test_data() -> Path:
    return TEST_DATA


@pytest.fixture
def config1(test_data: Path) -> BaseConfig:
    return load_config(test_data / "case1" / "config.toml")


@pytest.fixture
def config1_broken(test_data: Path) -> BaseConfig:
    return load_config(test_data / "case1" / "config-broken.toml")


def get_available_detectors() -> list[type[LicenseDetector]]:
    # TODO(anyone): Allow enforcing "strict mode" if any detectors are missing
    # This can be a env var and then enabled in the noxfile.
    available, _missing = gd({}, CONFIG1["licensing"])
    # HACK: We initialize the classes using a test config to check if they are
    # available and then return the base class so that it can be reinitialized
    return [type(d) for d in available.values()]


@pytest.fixture(name="detector", params=get_available_detectors())
def get_detectors(request) -> type[LicenseDetector]:
    return request.param



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
General configuration
"""

from __future__ import annotations

import os.path
from typing import Any, TypedDict, cast


class GeneralConfig(TypedDict):
    go_mod_dir: str | None


def create_general_config(config: dict[str, Any] | None = None) -> GeneralConfig:
    # Keep the same order here as in ArchiveConfig definition
    config = {} if config is None else config.copy()
    if (go_mod_dir := config.setdefault("go_mod_dir", None)) and os.path.isabs(
        go_mod_dir
    ):
        raise ValueError(f"general.go_mod_dir: {go_mod_dir!r} must be a relative path")
    return cast(GeneralConfig, config)



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

import contextlib
import os
import shlex
import time
from collections.abc import Iterable, Iterator, Sequence
from glob import iglob
from pathlib import Path
from tempfile import TemporaryDirectory
from types import SimpleNamespace
from typing import cast

import nox

IN_CI = "JOB_ID" in os.environ or "CI" in os.environ
ALLOW_EDITABLE = os.environ.get("ALLOW_EDITABLE", str(not IN_CI)).lower() in (
    "1",
    "true",
)

PROJECT = "go_vendor_tools"
SPECFILE = "go-vendor-tools.spec"
LINT_SESSIONS = ("formatters", "codeqa", "typing")
LINT_FILES = (
    f"src/{PROJECT}",
    "tests/pytests",
    "noxfile.py",
    "contrib",
    *iglob("doc/man/*.py"),
)
INTEGRATION_PACKAGES = ("autorestic", "fzf")
HAS_SCANCODE = os.environ.get("NO_SCANCODE") != "true"
COVERAGE_FAIL_UNDER = os.environ.get("COVERAGE_FAIL_UNDER") or (
    "90" if HAS_SCANCODE else "89"
)

nox.options.sessions = ("lint", "covtest")
nox.options.error_on_external_run = True
os.environ.setdefault("__GVT_TEST_MACROS", "1")


# Helpers


def install(session: nox.Session, *args, editable=False, **kwargs):
    if editable and ALLOW_EDITABLE:
        args = ("-e", *args)
    session.install(*args, "-c", "constraints.txt", **kwargs)


def git(session: nox.Session, *args, **kwargs):
    return session.run("git", *args, **kwargs, external=True)


def get_test_deps() -> Iterable[str]:
    yield ".[test]"
    if HAS_SCANCODE:
        yield ".[scancode]"


BASE_COVERAGE_COMMAND = ("coverage", "run", "-p", "--source", "go_vendor_tools")
COVERAGE_COMMANDS = SimpleNamespace(
    go_vendor_archive=(
        *BASE_COVERAGE_COMMAND,
        "-m",
        "go_vendor_tools.cli.go_vendor_archive",
    ),
    go_vendor_license=(
        *BASE_COVERAGE_COMMAND,
        "-m",
        "go_vendor_tools.cli.go_vendor_license",
    ),
    gocheck2=(
        *BASE_COVERAGE_COMMAND,
        "-m",
        "go_vendor_tools.cli.gocheck2",
    ),
)


@contextlib.contextmanager
def coverage_run(session: nox.Session) -> Iterator[dict[str, str | None]]:
    tmp = Path(session.create_tmp())
    covfile = (tmp / ".coverage").resolve()
    cov_env = {
        "COVERAGE_FILE": str(covfile),
        "GO_VENDOR_ARCHIVE": shlex.join(COVERAGE_COMMANDS.go_vendor_archive),
        "GO_VENDOR_LICENSE": shlex.join(COVERAGE_COMMANDS.go_vendor_license),
        **session.env,
    }
    yield cov_env
    combined = map(str, tmp.glob(".coverage.*"))
    session.run("coverage", "combine", *combined, env=cov_env)


# General


@nox.session(python=["3.9", "3.10", "3.11", "3.12", "3.13"])
def test(session: nox.Session):
    packages: list[str] = [*get_test_deps()]
    env: dict[str, str] = {}
    tmp = Path(session.create_tmp())

    if any(i.startswith("--cov") for i in session.posargs):
        packages.extend(("coverage[toml]", "pytest-cov"))
        env["COVERAGE_FILE"] = str(tmp / ".coverage")

    install(session, *packages, editable=True)
    session.run("pytest", "src", "tests/pytests", *session.posargs, env=env)


@nox.session
def integration(session: nox.Session) -> None:
    install(session, *get_test_deps(), "coverage[toml]", editable=True)
    packages_env = os.environ.get("PACKAGES")
    packages = shlex.split(packages_env) if packages_env else INTEGRATION_PACKAGES
    script_dir = Path("contrib").resolve()
    with (
        coverage_run(session) as cov_env,
        # askalono does not like that the temporary directory is in the
        # gitignore'd .nox/ directory
        TemporaryDirectory(dir=os.environ.get("TMPDIR")) as tmp,
    ):
        cov_env["TMPDIR"] = str(tmp)
        for package in packages:
            with session.chdir(Path("tests/integration", package)):
                for script_name in ("integration-archive.sh", "verify-license.sh"):
                    for try_path in (Path.cwd(), script_dir):
                        if (script_path := try_path / script_name).exists():
                            break
                    session.run(
                        "bash",
                        "-x",
                        str(script_path),
                        env=cov_env,
                        external=True,
                    )


@nox.session(name="integration-test-build")
def integration_test_build(session: nox.Session):
    temp = Path(session.create_tmp()).resolve()
    temp /= str(round(time.time()))
    temp.mkdir()
    install(session, *get_test_deps(), "coverage[toml]", editable=True)
    packages_env = os.environ.get("PACKAGES")
    packages = shlex.split(packages_env) if packages_env else INTEGRATION_PACKAGES
    with coverage_run(session) as cov_env:
        coverage_command = shlex.join(COVERAGE_COMMANDS.go_vendor_license)
        gocheck2_coverage_command = shlex.join(COVERAGE_COMMANDS.gocheck2)
        assert coverage_command[0]
        rpm_eval = Path("rpmeval.sh").resolve()
        for package in packages:
            package_tmp = temp / package
            package_tmp.mkdir()
            with session.chdir(Path("tests/integration/", package)):
                # fmt: off
                session.run(
                    "bash",
                    "-x",
                    str(rpm_eval),
                    "-D", f"__go_vendor_license {coverage_command}",
                    "-D", f"__gocheck2 {gocheck2_coverage_command}",
                    "-D", f"_specdir {Path.cwd()}",
                    "-D", f"_sourcedir {Path.cwd()}",
                    "-D", f"_rpmdir {package_tmp}",
                    "--nodeps",
                    "-ba", f"{package}.spec",
                    env=cov_env|{"RPM": "rpmbuild"},
                    external=True,
                )
                # fmt: on
                post_build = Path("integration-post-build.sh")
                if post_build.is_file():
                    files = session.run(
                        "find",
                        package_tmp,
                        "-name",
                        "*.rpm",
                        silent=True,
                        external=True,
                    )
                    file_list = str(files).splitlines()
                    print("Got:", file_list)
                    session.run("bash", "-x", post_build, *file_list, external=True)


@nox.session
def coverage(session: nox.Session):
    install(session, "coverage[toml]")
    session.run("coverage", "combine", "--keep", *iglob(".nox/*/tmp/.coverage"))
    session.run("coverage", "xml")
    session.run("coverage", "html")
    session.run("coverage", "report", "--fail-under", COVERAGE_FAIL_UNDER)


@nox.session()
def covtest(session: nox.Session):
    session.run("rm", "-f", *iglob(".nox/*/tmp/.coverage*"), external=True)
    test_sessions = (f"test-{v}" for v in cast("Sequence[str]", test.python))
    for target in test_sessions:
        session.notify(target, ["--cov"])


@nox.session(name="all")
def all_(session: nox.Session):
    lint(session)
    covtest(session)
    session.notify("integration")
    session.notify("integration-test-build")
    session.notify("coverage")


@nox.session(venv_backend="none")
def lint(session: nox.Session):
    """
    Run formatters, codeqa, and typing sessions
    """
    for notify in LINT_SESSIONS:
        session.notify(notify)


@nox.session
def codeqa(session: nox.Session):
    install(session, ".[codeqa]")
    session.run("ruff", "check", *session.posargs, *LINT_FILES)
    session.run("shellcheck", *iglob("contrib/*.sh"))
    # fmt: off
    session.run(
        "pymarkdownlnt", "scan",
        *iglob("*.md"),
        *iglob("doc/*.md"),
        "doc/man/rpm_macros.md",
    )
    # fmt: on
    session.run("reuse", "lint")


@nox.session
def formatters(session: nox.Session):
    install(session, ".[formatters]")
    posargs = session.posargs
    if IN_CI:
        posargs.append("--check")
    session.run("black", *posargs, *LINT_FILES)
    session.run("isort", *posargs, *LINT_FILES)


@nox.session
def typing(session: nox.Session):
    install(session, ".[typing]", "-r", "contrib/requirements.txt", editable=True)
    session.run("mypy", *LINT_FILES)


@nox.session
def bump(session: nox.Session):
    version = session.posargs[0]

    install(session, "build", "releaserr >= 0.1.dev123", "fclogr")
    session.run("releaserr", "--version")

    session.run("releaserr", "check-tag", version)
    session.run("releaserr", "ensure-clean")
    session.run("releaserr", "set-version", "-s", "file", version)

    install(session, ".")

    # Bump specfile
    # fmt: off
    session.run(
        "fclogr", "bump",
        "--new", version,
        "--comment", f"Release {version}.",
        SPECFILE,
    )
    # fmt: on

    # Bump changelog, commit, and tag
    git(session, "add", SPECFILE, f"src/{PROJECT}/__init__.py")
    session.run("releaserr", "clog", version, "--tag")
    session.run("releaserr", "build", "--backend", "generic", "--isolated")


@nox.session
def publish(session: nox.Session):
    # Setup
    install(session, "twine", "releaserr")
    session.run("releaserr", "--version")

    session.run("releaserr", "ensure-clean")

    # Upload to PyPI
    session.run("releaserr", "upload")

    # Post-release bump
    session.run("releaserr", "post-version", "-s", "file")
    git(session, "add", f"src/{PROJECT}/__init__.py")
    git(session, "commit", "-S", "-m", "Post-release version bump")


@nox.session
def srpm(session: nox.Session, posargs=None):
    install(session, "fclogr")
    posargs = posargs or session.posargs
    session.run("fclogr", "--debug", "dev-srpm", *posargs, SPECFILE)


@nox.session
def mockbuild(session: nox.Session):
    tmp = Path(session.create_tmp())
    srpm(session, ("-o", tmp, "--keep"))
    margs = [
        "mock",
        "--spec",
        str(Path(tmp, SPECFILE)),
        "--source",
        str(tmp),
        *session.posargs,
    ]
    if not session.interactive:
        margs.append("--verbose")
    session.run(*margs, external=True)


@nox.session
def releaserr(session: nox.Session):
    session.install("releaserr")
    session.run("releaserr", *session.posargs)


@nox.session
def mkdocs(session: nox.Session) -> None:
    install(session, "-e", ".", "-r", "docs-requirements.txt")
    session.run("mkdocs", *(session.posargs if session.posargs else ["build"]))
